#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

try:
  import google.generativeai as genai
except:
  pass

try:
  import openai
except:
  pass

try:
  import mistralai
  from mistralai import client as mistral_client
except:
  pass

class AI:
  def __init__(self, api_key=None):
    self.api_key = api_key

  def SetAPIKey(self, api_key):
    self.api_key = api_key

  def __call__(self, prompt):
    raise NotImplementedError

  def InitFromSystemVariable(self):
    key = self.SystemAPIKey()
    assert key, 'Can not initialize %s as system does not have a key.' % self
    self.SetAPIKey(key)

  @classmethod
  def SystemAPIKey(cls):
    return os.getenv(cls.api_key_system_variable)

  @classmethod
  def Options(cls):
    yield GoogleGenAI
    yield OpenAI
    yield MistralAI

  @classmethod
  def Get(cls):
    system_vars = []
    for option in cls.Options():
      key = option.SystemAPIKey()
      system_vars.append(option.api_key_system_variable)
      if key:
        return option(key)
    assert False, 'Can not initialize AI. None of the AI API keys were set: %s' % system_vars
  
  def CutOffChatter(self, response):
    useful_start = response.index('{')
    useful_end = response.rfind('}') + 1
    useful_response = response[useful_start:useful_end]
    if response != useful_response:
      print('Cutting of chatter from:', response)
      print('Obtaining:', useful_response)
    return useful_response


class GoogleGenAI(AI):
  configured_api_key = None
  api_key_system_variable = 'LOGICLM_GOOGLE_GENAI_API_KEY'

  def __call__(self, prompt):
    if self.configured_api_key != self.api_key:
      genai.configure(api_key=self.api_key)
      self.configured_api_key = self.api_key
    model = genai.GenerativeModel('gemini-1.5-pro-latest')
    content = model.generate_content(
      prompt,
      generation_config=dict(
        max_output_tokens=512,
        temperature=0.2
      ))
    return self.CutOffChatter(content.text)

class OpenAI(AI):
  configured_api_key = None
  api_key_system_variable = 'LOGICLM_OPENAI_API_KEY'
  
  def __call__(self, prompt):
    client = openai.OpenAI(api_key=self.api_key)
    response = client.chat.completions.create(
      model="gpt-4-1106-preview",
      messages=[
        {
          "role": "user",
          "content": prompt
        }
      ],
      temperature=1,
      max_tokens=512,
      top_p=1,
      frequency_penalty=0,
      presence_penalty=0
    )
    return self.CutOffChatter(response.choices[0].message.content)

class MistralAI(AI):
  configuration_api_key = None
  api_key_system_variable = 'LOGICLM_MISTRALAI_API_KEY'

  def __call__(self, prompt):
    client = mistral_client.MistralClient(api_key=self.api_key)
    message = mistralai.models.chat_completion.ChatMessage(
      role="user", content=prompt)
    chat_response = client.chat(model='mistral-medium',
                                messages=[message])
    result = self.CutOffChatter(chat_response.choices[0].message.content)
    result = result.replace('\\_', '_')
    return result


def GetPromptTemplate(config):
  def MaybeDescription(call_object):
    if 'description' in call_object:
      return ': %s' % call_object['description']
    return ''

  params_str = lambda parameters: ', '.join(p['field_name']
                                             + ':' for p in parameters)
  result_lines = []
  result_lines.append('Please write configuration for an OLAP request.')
  result_lines.append('Available measures are:')
  result_lines.extend('* %s(%s)%s' % (m['aggregating_function']['predicate_name'],
                                    params_str(m['aggregating_function'].get('parameters', [])),
                                    MaybeDescription(m))
                      for m in config['measures'])
  result_lines.append('')
  result_lines.append('Available dimensions are:')
  result_lines.extend('* %s(%s)%s' % (d['function']['predicate_name'],
                                    params_str(d['function'].get('parameters', [])),
                                    MaybeDescription(d))
                      for d in config['dimensions'])
  result_lines.append('')
  result_lines.append('Available filters are:')
  result_lines.extend('* %s(%s)%s' % (d['predicate']['predicate_name'],
                                      params_str(d['predicate'].get('parameters', [])),
                                      MaybeDescription(d))
                      for d in config['filters'])
  result_lines.extend('* %s(%s)%s' % (c['predicate']['predicate_name'],
                                      params_str(c['predicate'].get('parameters', [])),
                                      MaybeDescription(c))
                      for c in config['chart_types'])
  result_lines.append('Config is JSON object with fields title, measures, dimensions, filters, order, limit and chartType.')
  # result_lines.append('In the order clause please back-tick the predicate call.')
  result_lines.append('Always use all the fields. For example if you do not have filters, then pass it as empty list.')
  result_lines.append('')
  result_lines.extend(config['suffix_lines'])
  result_lines.append('')
  result_lines.append('Write me JSON for this request: __USER_REQUEST__')
  return '\n'.join(result_lines)


if __name__ == '__main__':
  ai = AI.Get()
  print(ai(sys.argv[1]))





#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import glob
import json
import logiclm
import io
import sys
from logica.common import color

def ShowFirstDifference(a, b):
  lines_a = a.split('\n')
  lines_b = b.split('\n')
  lines_a += ['END OF INPUT']
  lines_b += ['END OF INPUT']
  for i, (line_a, line_b) in enumerate(zip(lines_a, lines_b)):
    if line_a != line_b:
      print(color.Warn('Common start:'))
      print('\n'.join(lines_a[:i]))
      print(color.Warn('Actual line') + ' %d: %s' % (i, line_a))
      print(color.Warn('Golden line') + ' %d: %s' % (i, line_b))
      break
  else:
    print('A and B are the same.')


def RunTest(test_file, golden_run):
  print('%-70s %s' % (test_file, color.Format('{warning}RUNNING{end}')))
  with open(test_file) as f:
    content = f.read()
  test_config, golden = content.split('\n-----\n')
  test_config = json.loads(test_config)
  argv = ['',
          test_config['config'],
          test_config['command']] + (
            [json.dumps(test_config['request'])]
            if 'request' in test_config else [])
  true_stdout = sys.stdout
  mock_stdout = io.StringIO()
  sys.stdout = mock_stdout
  logiclm.main(argv)
  sys.stdout = true_stdout
  result = mock_stdout.getvalue()

  print('\033[F', end='')
  if result == golden:
    print('%-70s %s' % (test_file, color.Format('{ok}PASS   {end}')))
  else:
    print('%-70s %s' % (test_file, color.Format('{error}FAIL   {end}')), end='\r')
    if golden_run:
      with open(test_file, 'w') as w:
        w.write(json.dumps(test_config, indent=2))
        w.write('\n-----\n')
        w.write(result)
      print('%-70s %s' % (test_file, color.Format('{warning}UPDATED{end}')))
    else:
      print('')
      ShowFirstDifference(result, golden)


test_files = glob.glob('test_data/integration_tests/*.txt')
for test_file in test_files:
  golden_run = 'golden_run' in sys.argv
  RunTest(test_file, golden_run)


#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import json

import olap
import ai
import server


def Understand(config, user_request):
  mind = ai.AI.Get()
  template = ai.GetPromptTemplate(config)
  json_str = mind(template.replace('__USER_REQUEST__', user_request))
  try:
    json_obj = json.loads(json_str)
  except Exception as e:
    print('Failed parsing:', json_str)
    raise e
  return json_obj


def main(argv):
  config_filename = argv[1]
  command = argv[2]
  with open(config_filename) as f:
    config = json.loads(f.read())

  if command == 'understand':
    user_request = argv[3]
    print(Understand(config, user_request))
  elif command == 'logic_program':
    request = json.loads(argv[3])
    analyzer = olap.Olap(config, request)
    print(analyzer.GetLogicProgram())
  elif command == 'sql':
    request = json.loads(argv[3])
    analyzer = olap.Olap(config, request)
    print(analyzer.GetSQL())
  elif command == 'show_prompt':
    print(ai.GetPromptTemplate(config))
  elif command == 'understand_and_program':
    user_request = argv[3]
    request = Understand(config, user_request)
    analyzer = olap.Olap(config, request)
    print(analyzer.GetLogicProgram())
  elif command == 'understand_and_sql':
    user_request = argv[3]
    request = Understand(config, user_request)
    analyzer = olap.Olap(config, request)
    print(analyzer.GetSQL())
  elif command == 'start_server':
    server.StartServer(config)
  elif command == 'remove_dashboard_from_config':
    config['dashboard'] = {}
    print(json.dumps(config, indent='  '))
  else:
    assert False


if __name__ == '__main__':
  main(sys.argv)


#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import cgi
import json
from http import server
import socketserver
import time
import os
import tempfile
from urllib import parse
import ai
import olap
from logica.tools import run_in_terminal
from logica.parser_py import parse as parse_logica
from logica.compiler import rule_translate
from logica.type_inference.research import infer
from logica.common import sqlite3_logica

from logica.common import color
import io


class LogicLMServerHeart:
  def __init__(self, config):
    self.request_counter = 0
    self.nous = ai.AI.Get()
    self.prompt_template = ai.GetPromptTemplate(config)
    self.config = config
    color.CHR_ERROR = '<span style="color:red;">'
    color.CHR_END = '</span>'
    color.CHR_WARNING = '<span style="font-weight: bold">'
    color.CHR_UNDERLINE = '<span style="font-weight: bold;">'

  def StaticFilename(self, filename):
    return os.path.dirname(__file__) + '/html/' + filename

  def SpliceIn(self, template, cut_line, content):
    decorated_cut_line = '\n%s\n' % cut_line
    up, unused_cut_out, down = template.split('%s' % decorated_cut_line)
    return decorated_cut_line.join([up, content, down])

  def RawHtml(self):
    with open(self.StaticFilename('index.html')) as html_file:
      return html_file.read()
  
  def Html(self):
    html = self.RawHtml()
    html = self.SpliceIn(
        html,
        '  // < - dashboard cut out line. do not modify - >',
        'let tmpDashboardMappings = ' +
        json.dumps(self.config['dashboard'], indent=2),
    )
    html = self.SpliceIn(
        html,
        '  // < - intelligence config cut out line. do not modify. - >',
        'let intelligenceConfig = ' +
        json.dumps(self.LegacyIntelligenceConfig(), indent=2),
    )
    if 'custom_header' in self.config:
      html = self.SpliceIn(
        html,
        '    <!-- header title cut. do not modify. -->',
        self.config['custom_header']
      )
    if 'sample_questions' in self.config:
      html = html.SpliceIn(
        html,
        '  // < - sample questions cut. do not modify. - >',
        'let '
      )
    if 'tagline' in self.config:
      html = html.replace('The only true wisdom is in knowing you know nothing.',
                          self.config['tagline'])
    if 'example_question' in self.config:
      html = html.replace('Enter your request here.',
                          self.config['example_question'])
    return html

  def LogoPng(self):
    with open(self.StaticFilename('logiclm.png'), 'rb') as logo_file:
      return logo_file.read()

  # TODO: Refactor this.
  def LegacyIntelligenceConfig(self):
    intelligence_config = {}
    intelligence_config['aiDataRequest'] = {}
    intelligence_config['aiDataRequest']['dimensions'] = [
        {'functionName': d['function']['predicate_name']}
        for d in self.config['dimensions']]
    intelligence_config['aiDataRequest']['measures'] = [
        {'aggregatingFunctionName': d['aggregating_function']['predicate_name']}
        for d in self.config['measures']]
    intelligence_config['aiDataRequest']['filters'] = [
        {'predicateName': d['predicate']['predicate_name']}
        for d in self.config['filters']]
    intelligence_config['dashboard'] = bool(self.config['dashboard'])
    intelligence_config['aiVisualizationRequest'] = {
      'chartTypes': [
        {'chartTypeName': p['predicate']['predicate_name']}
        for p in self.config['chart_types']]
    }
    return intelligence_config

  def NaturalLanguageToRequestJson(self, user_request):
    json_request_str = self.nous(self.prompt_template.replace('__USER_REQUEST__',
                                                              user_request))
    print('AI response:', json_request_str)
    json_request = json.loads(json_request_str)
    json_request['exampleQuery'] = user_request
    # TODO: Change HTML to understand raw config.
    json_request['intelligence_config'] = self.LegacyIntelligenceConfig()
    return json_request
  
  def RunJson(self, json_request):
    if len(json_request['dimensions']) < 1 or len(json_request['measures']) < 1:
      json_request['nice_error'] = '<i>Please specify at least one measure and at least one dimension.</i>'
      return 'Fail(true)', "select 'fail'", []

    o = olap.Olap(self.config, json_request)
    charting_call = o.AsPredicateCall(json_request['chartType'])
    json_request['chart_type_predicate_call'] = {
      'predicate_name': charting_call.predicate_name,
      'arguments': {k: v.AsJson() for k, v in charting_call.named_args.items()}
    }
    try:
      logic_program = str(o.GetFullLogicProgram())
    except parse_logica.ParsingException as e:
      print('Failure of parsing:')
      e.ShowMessage()
      s = io.StringIO()
      e.ShowMessage(stream=s)
      json_request['nice_error'] = s.getvalue()
      return 'Fail(true)', "select 'fail'", [] 

    print('Logic program:')
    print(logic_program)

    try:
      sql = o.GetSQL()
    except parse_logica.ParsingException as e:
      print('Failure of parsing when building SQL:')
      e.ShowMessage()
      s = io.StringIO()
      e.ShowMessage(stream=s)
      json_request['nice_error'] = s.getvalue()
      return 'Fail(true)', "select 'fail'", []
    except rule_translate.RuleCompileException as e:
      print('Failure of compilation when building SQL:')
      e.ShowMessage()
      s = io.StringIO()
      e.ShowMessage(stream=s)
      json_request['nice_error'] = s.getvalue()
      return 'Fail(true)', "select 'fail'", []
    except infer.TypeErrorCaughtException as e:
      print('Failure of typing when building SQL:')
      e.ShowMessage()
      s = io.StringIO()
      e.ShowMessage(stream=s)
      json_request['nice_error'] = s.getvalue()
      return 'Fail(true)', "select 'fail'", []

    with  tempfile.TemporaryDirectory('LogicLM') as temp_dir:
      temp_file = '%s/report.l' % temp_dir
      with open(temp_file, 'w') as w:
        w.write(logic_program)
      header, rows = run_in_terminal.Run(temp_file, 'Report', output_format='header_rows')
    data = [header] + rows
    print('Data:', data)
    print(sqlite3_logica.ArtisticTable(header, rows))
    return logic_program, sql, data


def MakeSimpleLogicLMServer(config):
  heart = LogicLMServerHeart(config)
  class SimpleLogicLMServer(server.SimpleHTTPRequestHandler):
    def __init__(self, request, client_address, server):
      self.heart = heart
      super().__init__(request, client_address, server)

    def do_POST(self) -> None:
      url = parse.urlparse(self.path)
      ctype, pdict = cgi.parse_header(self.headers.get('content-type'))
      if url.path == '/understand_command':
        user_request = self.rfile.read(int(self.headers['Content-Length'])).decode('utf-8')
        print('User request:', user_request)
        json_request = self.heart.NaturalLanguageToRequestJson(user_request)
        print('LLM translation:', json.dumps(json_request, indent=' '))
        self.send_response(200)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(bytes(json.dumps(json_request), 'utf8'))
      if url.path == '/execute_config':
        json_request = json.loads(
          self.rfile.read(int(self.headers['Content-Length'])).decode('utf-8'))
        print('JSON request:', json_request)
        logic_program, sql, data = self.heart.RunJson(json_request)
        response = json_request | {
          'data': data,
          'sql': sql,
          'logical_program': logic_program,
        }
        self.send_response(200)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(bytes(json.dumps(response), 'utf8'))

    def do_GET(self) -> None:
      url = parse.urlparse(self.path)
      supported_paths = ['/index.html', '/logiclm.png']
      if url.path not in supported_paths:
        path = '/index.html'
      else:
        path = url.path
      if path == '/index.html':
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(bytes(self.heart.Html(), 'utf8'))
        return
      if path == '/logiclm.png':
        self.send_response(200)
        self.send_header('Content-type', 'image/png')
        self.end_headers()
        self.wfile.write(self.heart.LogoPng())
        return

      self.send_response(200)
      self.send_header("Content-type", "text/html")
      self.end_headers()
      self.wfile.write(bytes('Congratulations! You achieved impossible!', 'utf8'))
      self.heart.request_counter += 1

  return SimpleLogicLMServer


class ThreadedTCPServer(socketserver.ThreadingMixIn, 
                        socketserver.TCPServer):
    allow_reuse_address = True  # To avoid waiting after killing server.
    pass


def StartServer(config):
  simple_server = MakeSimpleLogicLMServer(config)
  port = config.get('port', 1791)
  server_instance = ThreadedTCPServer(('localhost', port), simple_server)
  print('Starting LogicLM server for "%s" intelligence configuration.' % config['name'])
  try:
    server_instance.serve_forever()
  except KeyboardInterrupt:
    print('Server terminated with Ctrl-C')
  finally:
    server_instance.server_close()


if __name__ == '__main__':
  config = json.loads(open('examples/baby_names/baby_names.json').read())
  StartServer(config)



#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jsonschema


def String():
  return {'type': 'string'}

def Number():
  return {'type': 'number'}

def Object(properties):
  return {'type': 'object',
          'properties': properties}

def FieldDescription():
  return Object({
    'field_name': String(),
    # TODO: Add field type.
  })

def PredicateSignature():
  return Object({
    'predicate_name': String(),
    'parameters': List(FieldDescription())
  })

def List(x):
  return {
    'type': 'array',
    'items': x
  }

def Measure():
  return Object({
    'aggregating_function': PredicateSignature(),
    'fact_table': String()
  })

def Dimension():
  return Object({
    'function': PredicateSignature()
  })

def Filter():
  return Object({
    'predicate': PredicateSignature(),
    'depends_on_dimensions': List(String())
  })

def NamedDimension():
  return Object({
    'name': String(),
    'dimension': String()
  })

def FactTable():
  return Object({
    'fact_table': String(),
    'consolidation': Object({
      'consolidated_fact_table': String(),
      'projected_dimensions': List(NamedDimension()),
      'consolidated_dimensions': List(NamedDimension()),
    }),
    'union': Object({
      'fact_tables': List(String())
    }),
    'ephemeral_dimensions': List(String()),
    'hostile_dimensions': List(String()),
  })

def OlapConfig():
  return Object({
    'name': String(),
    'fact_tables': List(FactTable()),
    'default_fact_table': String(),
    'measures': List(Measure()),
    'dimensions': List(Dimension()),
    'filters': List(Filter())
  })



#!/usr/bin/python
#
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import jsonschema
import json
import schema
import sys

from logica.compiler import universe
from logica.compiler import rule_translate
from logica.tools import avatar
from logica.parser_py import parse

class Olap:
  def __init__(self, config, request):
    jsonschema.validate(config, schema.OlapConfig())
    self.config = config
    self.request = request
    self.called_predicate_cache = {}
    self.measures = request.get('measures', [])
    default_fact_table = config['default_fact_table']
    self.fact_table_of_measure = {
      m['aggregating_function']['predicate_name']: m.get('fact_table',
                                                         default_fact_table)
      for m in self.config['measures']}
    
    self.default_fact_table = default_fact_table
    self.dimensions = request.get('dimensions', [])
    self.filters = request.get('filters', [])
    self.limit = request.get('limit', -1)
    self.order = request.get('order', [])
    self.direct_dependency = self.BuildDirectFactualDependencies()
    self.fact_dependencies = self.BuildFactualDependencies()
    self.fact_table_to_measures = self.BuildFactTableToMeasures()
    self.consolidation_info = self.GetConsolidationInfo()
    self.union_info = self.GetUnionInfo()
    self.table_needed_by_measure = {
        m: self.fact_table_of_measure[self.CalledPredicate(m)]
        for m in self.measures}
    self.measures_to_compute_from_table = self.BuildMeasuresToComputeFromTable()
    self.relevant_fact_tables = self.BuildListOfAllNeededTables()
    self.table_to_ephemeral_dimensions = self.BuildEphemeralDimensions()
    self.filter_to_needed_dimensions = self.BuildFilterToNeededDimensions()

  def DimensionsDomainRule(self) -> avatar.Rule:
    dimensions_domain_predicate = avatar.Predicate('DimensionsDomain')
    fact_variable = avatar.Variable('fact')
    dimensions_args = {
        self.ColumnName(d): self.AsPredicateCall(d)(fact_variable)
        for d in self.dimensions
    }
    head = dimensions_domain_predicate(**dimensions_args)
    filters_proposition = avatar.Conjunction([
      self.AsPredicateCall(f)(fact_variable) for f in self.filters])
    body = filters_proposition & avatar.Predicate(self.default_fact_table)(fact_variable)
    return +head << body

  def BuildFilterToNeededDimensions(self):
    result = {}
    for f in self.config['filters']:
      result[f['predicate']['predicate_name']] = f.get('depends_on_dimensions', [])

    return result

  def BuildEphemeralDimensions(self):
    result = {}
    for t in self.config['fact_tables']:
      result[t['fact_table']] = t.get('ephemeral_dimensions', [])
    return result

  def BuildListOfAllNeededTables(self):
    result = set()
    for t in self.measures_to_compute_from_table:
      result |= {t}
      result |= set(self.fact_dependencies[t])
    return list(sorted(result))

  def BuildMeasuresToComputeFromTable(self):
    result = {}
    for measure, table in self.table_needed_by_measure.items():
      result[table] = result.get(table, []) + [measure]
    return result

  def BuildFactTableToMeasures(self):
    result = {}
    for m in self.request['measures']:
      measure = self.CalledPredicate(m)
      table = self.fact_table_of_measure[measure]
      result[table] = result.get(table, []) + [measure]
    return result
  
  def GetConsolidationInfo(self) -> dict[str, tuple[str,
                                                    dict[str, str],
                                                    dict[str, str]]]:
    result = {}
    for f in self.config['fact_tables']:
      name = f['fact_table']
      if consolidation := f.get('consolidation'):
        result[name] = (consolidation.get('consolidated_fact_table'),
                        consolidation.get('consolidated_dimensions', []),
                        consolidation.get('projected_dimensions', []))
    return result

  def GetUnionInfo(self):
    result = {}
    for f in self.config['fact_tables']:
      name = f['fact_table']
      if union := f.get('union'):
        result[name] = (union['fact_tables'],
                        union.get('consolidated_dimensions', []),
                        union.get('projected_dimensions', []))
    return result
                        
  def BuildDirectFactualDependencies(self):
    direct_dependency = {}
    all_fact_tables = []
    for f in self.config['fact_tables']:
      all_fact_tables.append(f['fact_table'])
      if 'consolidation' in f:
        direct_dependency[f['fact_table']] = {
          f['consolidation']['consolidated_fact_table']}
      if 'union' in f:
        direct_dependency[f['fact_table']] = set(f['union']['fact_tables'])
    self.all_fact_tables = all_fact_tables
    for k, v in direct_dependency.items():
      assert k in all_fact_tables, (k, v)
      assert v <= set(all_fact_tables), ((k, v), all_fact_tables)


    return direct_dependency

  def BuildFactualDependencies(self):
    direct_dependency = self.direct_dependency
    all_fact_tables = self.all_fact_tables

    all_dependencies = {}
    def GetAllDependencies(t):
      if t not in direct_dependency:
        return set()
      if t not in all_dependencies:
        all_dependencies[t] = direct_dependency[t]
        for x in direct_dependency[t]:
          all_dependencies[t] = (
            all_dependencies[t] |
            GetAllDependencies(x))
      return all_dependencies[t]
    for t in all_fact_tables:
      all_dependencies[t] = GetAllDependencies(t)
    return {t: sorted(all_dependencies[t])
            for t in all_fact_tables}

  def OldLogicProgram(self, request):
    Report = avatar.Predicate('Report')
    fact = avatar.Variable('fact')

    head = +Report(**{f'`{measure}`': avatar.Predicate(measure)(fact)
                      for measure in request['measures'] })

    fact_table = avatar.Predicate(self.config['fact_tables'][0]['fact_table'])

    body = fact_table(fact)
    rule = head << body
    return str(rule)

  def ColumnName(self, predicate_call):
    p = self.CalledPredicate(predicate_call)
    disambiguation = abs(Hash(predicate_call)) % 1000000
    return '_'.join([p.lower(), str(disambiguation)])

  def UnionFacts(self, unioned_table, component_fact_tables):
    unioned_predicate = avatar.Predicate(unioned_table)

    fact_variable = avatar.Variable('fact')
    head = unioned_predicate(fact_variable)
    disjuncts = [avatar.Predicate(t)(fact_variable)
                 for t in component_fact_tables]
    body = avatar.Disjunction(disjuncts)
    return head << body

  def WrapFacts(self, resulting_fact_table,
                input_rule: avatar.Rule,
                dimensions_domain_rule: avatar.Rule):
    head_args = {}
    domain_call_args = {}
    for domain_arg in dimensions_domain_rule.head.named_args.keys():
      domain_call_args[domain_arg] = head_args[domain_arg] = avatar.Variable(domain_arg)

    for input_arg in input_rule.head.named_args.keys():
      head_args[input_arg] = avatar.Variable(input_arg)

    columns = input_rule.head.named_args.keys()
    input_predicate = input_rule.head.predicate_name
    args = {c: avatar.Variable(c) for c in columns}
    head = avatar.Predicate(resulting_fact_table)(avatar.Literal(head_args))
    body = avatar.Conjunction([
      avatar.Predicate(input_predicate)(**args),
      avatar.Predicate(dimensions_domain_rule.head.predicate_name)(**domain_call_args)
      ])
    return head << body

  def ConsolidateFacts(self, fact_table, measures, dimensions, filters,
                       consolidated_dimensions, projected_dimensions,
                       translucent_dimensions,
                       consolidating_predicate_name=None):
    """Produces rule and predicate that consolidates the fact table."""
    consolidating_predicate_name = consolidating_predicate_name or ('Consolidating' + fact_table)
    consolidating_predicate = avatar.Predicate(consolidating_predicate_name)
    fact_variable = avatar.Variable('fact')
    Aggr = lambda a : avatar.Aggregation('Aggr', a)
    measures_args = {
        self.ColumnName(m): Aggr(self.AsPredicateCall(m)(fact_variable))
        for m in measures}
    consolidated_dimensions_args = {
        k: Aggr(self.AsPredicateCall(v)(fact_variable))
        for k, v in consolidated_dimensions.items()
    }
    dimensions_args = {
        self.ColumnName(d): self.AsPredicateCall(d)(fact_variable)
        for d in dimensions
    }
    projected_dimensions_args = {
        k: self.AsPredicateCall(v)(fact_variable)
        for k, v in projected_dimensions.items()
    }
    translucent_dimensions_args = {
        self.ColumnName(d): avatar.Subscript(fact_variable,
                                             self.ColumnName(d))
        for d in translucent_dimensions
    }
    filters_list = [self.AsPredicateCall(f)(fact_variable) for f in filters]

    fact_table_call = avatar.Predicate(fact_table)(fact_variable)
    body = fact_table_call & avatar.Conjunction(filters_list)
    head_args = {**measures_args,
                 **consolidated_dimensions_args,
                 **dimensions_args,
                 **projected_dimensions_args,
                 **translucent_dimensions_args}
    head = consolidating_predicate(**head_args)
    rule = +head << body
    return consolidating_predicate_name, rule

  def ParseExpression(self, s):
    return parse.ParseExpression(parse.HeritageAwareString(s))

  def CalledPredicate(self, predicate_call):
    if predicate_call not in self.called_predicate_cache:
      parsed_call = self.ParseExpression(predicate_call)
      assert 'call' in parsed_call, parsed_call
      self.called_predicate_cache[predicate_call] = parsed_call['call']['predicate_name']
    return self.called_predicate_cache[predicate_call]

  def AsPredicateCall(self, predicate_call_str):
    parsed_call = self.ParseExpression(predicate_call_str)
    call = avatar.LogicalTerm.FromSyntax(parsed_call)
    return call

  def FactTableDimensions(self, t):
    return [
      d for d in self.dimensions
      if self.CalledPredicate(d) not in self.table_to_ephemeral_dimensions[t]]
  
  def GetLogicProgram(self):
    self.source_of_measure = {}
    needs_building = []
    # Computing each measure.
    rules_for_measures = []
    program = avatar.Program([])
    dimensions_domain_rule = self.DimensionsDomainRule()
    need_dimensions_domain = False
    for i, (fact_table, measures) in enumerate(self.measures_to_compute_from_table.items()):
      dimensions = [
        d for d in self.dimensions
        if self.CalledPredicate(d) not in self.table_to_ephemeral_dimensions[fact_table]]
      filters = [
        f for f in self.filters
        if not set(self.filter_to_needed_dimensions[self.CalledPredicate(f)]) & set(self.table_to_ephemeral_dimensions[fact_table])
      ]  
      if fact_table in self.direct_dependency:
        needs_building += [fact_table]
        translucent_dimensions, dimensions = dimensions, []
        filters = []
      else:
        translucent_dimensions = []

      consolidated_table, rule = self.ConsolidateFacts(
        fact_table, measures, dimensions, filters, {}, {}, translucent_dimensions)

      rules_for_measures.append(rule)
      program.AddRule(rule)
      if i == 0:
        program.rules[i].comment_before_rule = 'Computing all the measures.'
    # Computing auxiliary tables.
    i = 0
    while i < len(needs_building):
      fact_table_to_build = needs_building[i]
      if fact_table_to_build in self.consolidation_info:
        t, c, p = self.consolidation_info[fact_table_to_build]
        if t in self.direct_dependency:
          needs_building += [t]
        dimensions = self.FactTableDimensions(t)
        filters = [
          f for f in self.filters
          if not (set(self.filter_to_needed_dimensions[self.CalledPredicate(f)]) <= set(self.table_to_ephemeral_dimensions[t]))
        ]
        consolidated_table, rule = self.ConsolidateFacts(
          t, [], dimensions, filters, 
          {x['name']: x['dimension'] for x in c},
          {x['name']: x['dimension'] for x in p},
          translucent_dimensions=[],
          consolidating_predicate_name=fact_table_to_build + 'Step1')
        program.AddRule(rule)
        program.AddRule(self.WrapFacts(fact_table_to_build, rule, dimensions_domain_rule))
        need_dimensions_domain = True
      elif fact_table_to_build in self.union_info:
        ts, c, p = self.union_info[fact_table_to_build]
        for t in ts:
          if t in self.direct_dependency:
            needs_building += [t]
        dimensions = [
          d for d in self.dimensions
          if self.CalledPredicate(d) not in self.table_to_ephemeral_dimensions[fact_table_to_build]]
        filters = [
          f for f in self.filters
          if self.filter_to_needed_dimensions[self.CalledPredicate(f)] not in self.table_to_ephemeral_dimensions[fact_table_to_build]
        ]
        program.AddRule(self.UnionFacts(fact_table_to_build, ts))
      i += 1
    if need_dimensions_domain:
      program.AddRule(dimensions_domain_rule
                      )
    def ColumnName(predicate_call_str):
      return '`%s`' % predicate_call_str.replace('(', '<').replace(')', '>').replace('"', "'").replace('"', "'")
    # Assembling all the measures together.
    measures_args = {ColumnName(m): avatar.Variable(self.ColumnName(m))
                     for m in self.measures}
    dimensions_args = {ColumnName(d): avatar.Variable(self.ColumnName(d))
                       for d in self.dimensions}
    if self.limit >= 0:
      program.AddRule(avatar.Predicate('@Limit')(
        avatar.Literal('Report'), avatar.Literal(self.limit)) << None)
    if self.order:
      def DecorateOrder(s):
        for suffix in ['asc', 'desc']:
          if s.endswith(suffix):
            direction = suffix
            s = s.removesuffix(' ' + suffix)
            break
        else:
          direction = 'asc'
        return ColumnName(s) + ' ' + direction
      program.AddRule(avatar.Predicate('@OrderBy')(
        avatar.Literal('Report'), *map(lambda x: avatar.Literal(DecorateOrder(x)),
                                       self.order)) << None)
    head = avatar.Predicate('Report')(**(dimensions_args | measures_args))
    body = avatar.Conjunction([])
    for rule in rules_for_measures:
      p = rule.head.predicate_name
      named_args = {a: avatar.Variable(a) for a in rule.head.named_args}
      body = body & avatar.Predicate(p)(**named_args)
    rule = head << body
    rule.comment_before_rule = 'Assembling all the measures.'
    program.AddRule(rule)
    return program

  def GetFullLogicProgram(self):
    base_program = open(self.config['logica_program']).read()
    incremental_program = self.GetLogicProgram()
    program = base_program + ';\n' + str(incremental_program)
    return program

  def GetSQL(self):
    program = self.GetFullLogicProgram()
    rules = parse.ParseFile(program)['rule']
    logic_program = universe.LogicaProgram(rules)
    sql = logic_program.FormattedPredicateSql('Report')
    return sql

def Hash(s):
  return abs(int(hashlib.md5(str(s).encode()).hexdigest()[:16], 16) - (1 << 63))

if __name__ == '__main__':
  config = json.loads(open('test_data/abstract_model/abstract.json').read())
  request = {"measures": ["NumberOfBabies()", "Fact2Measure(x: 3)", "MeasureOverConsolidated()"],
            "dimensions": ["State()", "Year()", "Dim4()"],
            "filters": ["StateIn(states: [\"NY\", \"WA\"])"]}
  olap = Olap(config, request)
  # p, r = olap.ConsolidateFacts(
  #   'T', ['M1(a:1)', 'M2(b:2)'], ['D1(a:1, b:2)'], ['F1(x:1)', 'F2()'],
  #   {'cd': 'ConsolidatedDim1(a: 1)'},
  #   {'pd': 'ProjectedDim1(x: 5)'})
  # print(p)
  # print(r)
  # print(olap.fact_dependencies)
  program = olap.GetLogicProgram()
  print('Program:')
  print(program)
  # print(olap.LogicProgram(request))